/*
 * Copyright (c) 1991-2020 by STEP Tools Inc. 
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and
 * its documentation is hereby granted, provided that this copyright
 * notice and license appear on all copies of the software.
 * 
 * STEP TOOLS MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE
 * SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING
 * BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT. STEP TOOLS
 * SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A
 * RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
 * DERIVATIVES.
 *
 * Author: David Loffredo (loffredo@steptools.com)
 */


#ifdef _WIN32
#include <windows.h>
#endif

#include <IfcRoot.h>
#include "make_guid.h"

//----------------------------------------
// ID AND HISTORY UTILITIES -- Handle the owner history and global id
// fields required on all IfcRoot instances.  The id functions have
// been written to use the appopriate platform functions for Windows,
// Mac, and Linux, but you may need to add some code if you port to
// other platforms.
//


extern void assign_ifc_guid (IfcRoot * obj);


#ifdef _WIN32
//----------------------------------------
// WINDOWS, uuid from CoCreateGuid, link with ole32.lib
#define GETUUID	GetWindowsUUID

static void GetWindowsUUID (unsigned char uuid[16])
{
    GUID guid;
    CoCreateGuid (&guid);

    // Data1 Specifies the first 8 hexadecimal digits of the GUID.
    uuid[0] = (unsigned char) (guid.Data1 >> 24) & 0xff;
    uuid[1] = (unsigned char) (guid.Data1 >> 16) & 0xff;
    uuid[2] = (unsigned char) (guid.Data1 >> 8) & 0xff;
    uuid[3] = (unsigned char) (guid.Data1) & 0xff;

    uuid[4] = (unsigned char) (guid.Data2 >> 8) & 0xff;
    uuid[5] = (unsigned char) (guid.Data2) & 0xff;

    uuid[6] = (unsigned char) (guid.Data3 >> 8) & 0xff;
    uuid[7] = (unsigned char) (guid.Data3) & 0xff;

    for (unsigned i = 0; i<8; i++)
	uuid[8+i] = guid.Data4[i];
}    
#endif 

#ifdef __APPLE__
//----------------------------------------
// MACOS, uuid from CFUUIDCreate, link with -framework CoreFoundation
#include <CoreFoundation/CoreFoundation.h>
#define GETUUID GetMacUUID

static void GetMacUUID (unsigned char uuid[16])
{
    CFUUIDRef idref = CFUUIDCreate(NULL);
    CFUUIDBytes idbytes = CFUUIDGetUUIDBytes(idref);

    uuid[0] = idbytes.byte0;
    uuid[1] = idbytes.byte1;
    uuid[2] = idbytes.byte2;
    uuid[3] = idbytes.byte3;
    uuid[4] = idbytes.byte4;
    uuid[5] = idbytes.byte5;
    uuid[6] = idbytes.byte6;
    uuid[7] = idbytes.byte7;
    uuid[8] = idbytes.byte8;
    uuid[9] = idbytes.byte9;
    uuid[10] = idbytes.byte10;
    uuid[11] = idbytes.byte11;
    uuid[12] = idbytes.byte12;
    uuid[13] = idbytes.byte13;
    uuid[14] = idbytes.byte14;
    uuid[15] = idbytes.byte15;
    
    CFRelease (idref);
}
#endif


#if defined(linux) || defined(__linux__)
//----------------------------------------
// LINUX, uuid from uuid_generate(), link with -luuid.  On Linux, the
// uuid_t definition is also an array[16] of bytes, so we just use the
// library function directly.
// 
#include <uuid/uuid.h>
#define GETUUID	uuid_generate
#endif


#ifndef GETUUID
//----------------------------------------
// OTHER PLATFORMS, you will need to find the appropriate
// platform-specific function or get the ethernet id and make your
// own.
#define GETUUID	unknown_uuid_function

static void unknown_uuid_function (unsigned char uuid[16])
{
    // fill with random values as a temporary measure
    for (int i=0; i<16; i++) {
	uuid[i] = (unsigned char) (rand() % 0x100);
    }
}
#endif





// Base 64 encode a 128bit identifier using the character set and
// approach given in the IFC documentation.  Get the UUID from an OS
// specific function or build it yourself 
//
static void Base64EncodeUUID (const unsigned char uuid[16], char uuidstr[23])
{
    static char base64char[] =
	"0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz_$";

    char * str = uuidstr;

    // first byte do manually because the first char is a partial
    // value.  As per the sample code in the ifc documentation.
    //
    *str++ = base64char[(uuid[0] & 0xc0) >> 6];
    *str++ = base64char[(uuid[0] & 0x3f)];


    // Break the remaining 15 bytes into three bytes groups, then for
    // each, divide into four 6bit values.
    for (unsigned i=1; i<15; i+=3)
    {
	*str++ = base64char[(uuid[i] & 0xfc) >> 2];

	*str++ = base64char[
	    ((uuid[i] & 0x03) << 4) |
	    ((uuid[i+1] & 0xf0) >> 4)
	    ];

	*str++ = base64char[
	    ((uuid[i+1] & 0x0f) << 2) |
	    ((uuid[i+2] & 0xc0) >> 6)
	    ];

	*str++ = base64char[(uuid[i+2] & 0x3f)];
    }
    
    *str = 0;	// done
}    





void assign_ifc_guid (IfcRoot * obj)
{
    unsigned char uuid[16];
    char uuidstr[23];

    // Allocate a new UUID and assign a base64 encoded version to the
    // IFC root object.

    if (obj) {
	GETUUID (uuid);
	Base64EncodeUUID (uuid, uuidstr);
	obj-> GlobalId (uuidstr);
    }
}

